@@ -95,11 +95,15 @@ use tokio::io::AsyncWriteExt;
 use tokio::io::BufStream;
 use tokio::io::BufWriter;
 
+use tracing::instrument;
+
 mod error;
 pub use error::Error;
 pub use error::StreamError;
 pub mod stream;
 
+const CURRENT_FORMAT_VERSION: u16 = 1;
+
 fn codec() -> impl bincode::Options + Copy /* {{{ */ {
 	bincode::DefaultOptions::new()
 		.with_limit(u32::max_value() as u64)
@@ -138,6 +142,8 @@ impl PartialOrd for IndexEntry /* {{{ */ {
 
 #[derive(Debug, Deserialize, Serialize)]
 struct Header {
+	format_version: u16,
+	completed_writing: bool,
 	pair_count: u64,
 	/// absolute position in the file
 	index_start_pos: u64
@@ -168,6 +174,8 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 			path: path.as_ref().to_path_buf(),
 			file: BufStream::new(fd),
 			header: Header{
+				format_version: 0,
+				completed_writing: false,
 				pair_count: 0,
 				index_start_pos: Self::HEADER_SIZE as u64
 			},
@@ -184,9 +192,9 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 		Self::HEADER_SIZE as u64
 	} // }}}
 
+	#[instrument(err, level = "trace", skip(self))]
 	async fn get_header(&mut self) -> Result<Header, Error> /* {{{ */ {
 		let mut buf = [0u8; size_of::<Header>()];
-		//let mut buf = [0u8; 8];
 		self.file.seek(SeekFrom::Start(0)).await?;
 		self.file.read(&mut buf).await?;
 		Ok(deserialize(&buf)?)
@@ -198,6 +206,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 		Ok(deserialize(&buf)?)
 	} // }}}
 
+	#[instrument(err, level = "trace", skip(self))]
 	async fn get_index_entry(&mut self, i: u64) -> Result<IndexEntry, Error> /* {{{ */ {
 		let pos = self.index_start_position() + (i * Self::INDEX_ENTRY_SIZE as u64);
 		self.file.seek(SeekFrom::Start(pos)).await?;
@@ -205,6 +214,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 		Ok(entry)
 	} // }}}
 
+	#[instrument(err, level = "debug", skip(self))]
 	async fn find_index_entry(&mut self, key: u64) -> Result<Option<IndexEntry>, Error> /* {{{ */ {
 		let mut size = self.header.pair_count;
 		let mut left = 0;
@@ -243,6 +253,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 		Ok(Some(deserialize(&buf)?))
 	} // }}}
 
+	#[instrument(err, level = "trace", skip(self))]
 	async fn read_value_at(&mut self, position: u64) -> Result::<T, Error> /* {{{ */ {
 		self.file.seek(SeekFrom::Start(position)).await?;
 		match self.read_value().await? {
@@ -251,10 +262,24 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 		}
 	} // }}}
 
+	/// Reads the header and checks to see whether or not the store is valid; at this time, this
+	/// means making sure the format is fully up-to-date and writing completed.
+	#[instrument(err, level = "trace", skip(self))]
+	#[inline]
+	pub async fn is_valid(&mut self) -> Result<bool, Error> /* {{{ */ {
+		let header = self.get_header().await?;
+		Ok(
+			header.format_version == CURRENT_FORMAT_VERSION &&
+			header.completed_writing
+		)
+	} // }}}
+
 	/// Provides a [sequential stream](stream::ValueStream) of all values in the Store.  In
 	/// relational database parlance, this would be a tablescan.
 	// TODO:  Right now, nothing stops you from starting a stream then writing
 	//    to the Store.  This will, naturally, ruin everything.
+	#[instrument(err, level = "debug", skip(self))]
+	#[inline]
 	pub async fn stream_values(&self) -> Result<impl Stream<Item=Result<T, Error>>, Error> /* {{{ */ {
 		let mut file = tokio::fs::OpenOptions::new()
 			.read(true)
@@ -271,6 +296,8 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 	/// Looks up a single value by key.  Will attempt to cache data positions in the backing file
 	/// to minimize disk seeking.  Returns [Error::NotFound] when `key` does not exist in the
 	/// Store.
+	#[instrument(err, level = "debug", skip(self))]
+	#[inline]
 	pub async fn get(&mut self, key: u64) -> Result<T, Error> /* {{{ */ {
 		let position = match self.index_cache.get(&key) {
 			Some(v) => *v,
@@ -290,6 +317,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 	/// Looks up multiple values by key.  Will find (uncached) keys in sorted order to minimize
 	/// disk seeking back and forth.  Does _not_ return an error when `key` does not exist; calling
 	/// code should check for `Some` vs `None` in the resulting `HashMap` for fallibility.
+	#[instrument(err, level = "debug", skip(self))]
 	pub async fn get_many(&mut self, keys_in: &[u64]) -> Result<HashMap<u64, T>, Error> /* {{{ */ {
 		let mut results = HashMap::with_capacity(keys_in.len());
 		let mut keys = Vec::new();
@@ -317,6 +345,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 	} // }}}
 
 	// Order of operations:
+	//   * Write down the header, with completed_writing = false
 	//   * Seek to data_start_position
 	//   * For each entry in pairs:
 	//     * Compute the serialized size of this entry
@@ -325,7 +354,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 	//     * Write the buffer to the file
 	//   * Save current file position (as index_start_pos) and number of entries to the header
 	//   * Write the sorted index down to the file
-	//   * Seek to the beginning of the file and write down the header
+	//   * Seek to the beginning of the file and write down the header, with completed_writing = true
 	async fn write_to<S, W>(&mut self, mut input_stream: S, writer: &mut W) -> Result<(), StreamError<S::Error>> /* {{{ */
 	where
 		S: TryStream<Ok=(u64, T)> + Unpin,
@@ -335,6 +364,13 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 		let mut buf = vec![0u8; 1024];
 		let encoder = codec();
 
+		self.header.format_version = CURRENT_FORMAT_VERSION;
+		self.header.completed_writing = false;
+		encoder.serialize_into(&mut buf[0..], &self.header)?;
+		writer.seek(SeekFrom::Start(0)).await?;
+		writer.write(&buf[0..Self::HEADER_SIZE]).await?;
+		writer.flush().await?;
+
 		let mut pos = self.data_start_position();
 		let mut index = Vec::new();
 
@@ -364,6 +400,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 			writer.write(&buf[0..Self::INDEX_ENTRY_SIZE]).await?;
 		}
 
+		self.header.completed_writing = true;
 		encoder.serialize_into(&mut buf[0..], &self.header)?;
 		writer.seek(SeekFrom::Start(0)).await?;
 		writer.write(&buf[0..Self::HEADER_SIZE]).await?;
@@ -430,6 +467,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 	///		));
 	///	}
 	/// ```
+	#[inline]
 	pub async fn write<S>(&mut self, input_stream: S) -> Result<(), StreamError<S::Error>> /* {{{ */
 	where
 		S: TryStream<Ok=(u64, T)> + Unpin,
@@ -442,6 +480,7 @@ impl<T: Serialize + DeserializeOwned + std::fmt::Debug> Store<T> {
 
 	/// For cases where the incoming data stream is infallible, this allows you to avoid that
 	/// portion of error handling.
+	#[inline]
 	pub async fn write_infallible<S>(&mut self, input_stream: S) -> Result<(), Error> /* {{{ */
 	where
 		S: Stream<Item=(u64, T)> + Unpin
@@ -599,6 +638,47 @@ mod tests {
 		}
 	} // }}}
 
+	#[tokio::test]
+	async fn strings_reopen() /* {{{ */ {
+		// setup {{{
+		#[derive(Clone, Debug, Deserialize, Serialize, PartialEq)]
+		struct Struct {
+			number: u8,
+			name: String,
+			maybe_name: Option<String>
+		}
+		impl Struct {
+			fn new(i: u8) -> Self {
+				Self{
+					number: i,
+					name: english_numbers::convert_all_fmt(i as i64),
+					maybe_name: match i % 2 {
+						0 => Some(english_numbers::convert_long(i as i64, english_numbers::Formatting::none())),
+						_ => None
+					}
+				}
+			}
+		}
+		let values = (0..255_u8).map(|i| Struct::new(i)).collect::<Vec<_>>();
+		// }}}
+
+		let path = tempfile::NamedTempFile::new().unwrap().into_temp_path();
+		for _ in 1..4 {
+			let mut store: Store<Struct> = Store::new(&path).await.unwrap();
+			store.write_infallible(stream::iter(values.clone().into_iter().enumerate().map(|(i, v)| (i as u64, v)))).await.unwrap();
+
+			let mut rng = rand::thread_rng();
+			let mut indices = (0..values.len() as u64).collect::<Vec<u64>>();
+			for _ in 1..16 {
+				indices.shuffle(&mut rng);
+				for i in indices.iter() {
+					let result = store.get(*i).await.unwrap();
+					assert_eq!(result, values[*i as usize]);
+				}
+			}
+		}
+	} // }}}
+
 	#[tokio::test]
 	async fn smallstream() /* {{{ */ {
 		let values = (0..5_u8).map(|i| FixedWidth::new(i)).collect::<Vec<_>>();
@@ -640,6 +720,29 @@ mod tests {
 		tokio::fs::remove_file(path).await.unwrap();
 	} // }}}
 
+	#[tokio::test]
+	async fn stream_reopen() /* {{{ */ {
+		let mut values = (0..255_u8).map(|i| FixedWidth::new(i)).collect::<Vec<_>>();
+		values.extend((0..255_u8).map(|i| FixedWidth::new2(i)));
+
+		let path = tempfile::NamedTempFile::new().unwrap().keep().unwrap().1;
+		for iteration in 1..4 {
+			println!(">>> {}", iteration);
+			let mut store: Store<FixedWidth> = Store::new(&path).await.unwrap();
+			store.write_infallible(stream::iter(values.clone().into_iter().enumerate().map(|(i, v)| (i as u64, v)))).await.unwrap();
+
+			let mut stream = store.stream_values().await.unwrap().enumerate();
+			while let Some((i, result)) = stream.next().await {
+				let value = result.unwrap();
+				assert_eq!(value, values[i]);
+			}
+
+			assert_eq!(store.get(0).await.unwrap(), values[0]);
+			assert_eq!(store.get(1).await.unwrap(), values[1]);
+		}
+		tokio::fs::remove_file(path).await.unwrap();
+	} // }}}
+
 	#[tokio::test]
 	async fn fw_constant() /* {{{ */ {
 		let values = (0..5).map(|i| FixedWidth::new(i)).collect::<Vec<_>>();
@@ -750,6 +853,27 @@ mod tests {
 		}
 	} // }}}
 
+	#[tokio::test]
+	async fn fw_prime_53_reopen() /* {{{ */ {
+		let values = (0..53).map(|i| FixedWidth::new(i)).collect::<Vec<_>>();
+
+		let path = tempfile::NamedTempFile::new().unwrap().into_temp_path();
+		for _ in 1..2 {
+			let mut store: Store<FixedWidth> = Store::new(&path).await.unwrap();
+			store.write_infallible(stream::iter(values.clone().into_iter().enumerate().map(|(i, v)| (i as u64, v)))).await.unwrap();
+
+			let mut rng = rand::thread_rng();
+			let mut indices = (0..values.len() as u64).collect::<Vec<u64>>();
+			for _ in 1..16 {
+				indices.shuffle(&mut rng);
+				for i in indices.iter() {
+					let result = store.get(*i).await.unwrap();
+					assert_eq!(result, values[*i as usize]);
+				}
+			}
+		}
+	} // }}}
+
 	#[tokio::test]
 	async fn get_many() /* {{{ */ {
 		let mut values = (0..255_u8).map(|i| FixedWidth::new(i)).collect::<Vec<_>>();
